#include <Actor.h>
#include <tinyxml.h>
#include "Axis3D.h"

Actor::Actor()
{
	skelSet = false;
	defaultWindow = 300;
	matName = "";
	playing = false;
	blending = false;
	stillBlend = false;
	rep = NULL;
	heading = 0;
}

Actor::~Actor()
{
}


void 
Actor::Draw()
{
	if (!skelSet)
		return;
	Material::ApplyManagedMaterial(matName.c_str());
	
}



bool 
Actor::LoadXML(const char * name)
{
	TiXmlDocument doc;
	if (!doc.LoadFile(name))
	{
		printf("Error loading actor file: %s\n",doc.ErrorDesc());
		return false;
	}
	
	TiXmlNode* node = 0;
	TiXmlHandle docHandle( &doc );
	
	TiXmlElement* skelEl = docHandle.FirstChild( "actor" ).FirstChild( "skeleton" ).Element();
	TiXmlElement* repEl = docHandle.FirstChild( "actor" ).FirstChild( "repetoire" ).Element();
	TiXmlElement* skinEl = docHandle.FirstChild( "actor" ).FirstChild( "skin" ).Element();


	if (!skelEl)
	{
		printf("Bad XML format - no skeleton node\n");
		return false;
	}

	const char * skelname = skelEl->Attribute("name");
	if (!skelname)
	{
		printf("Bad XML format - no skeleton name\n");
		return false;
	}

	if (!SkelChar::Load(skelname,skelname))
	{
		printf("Error loading skeleton");
		return false;
	}
	SetSkel(skelname);	



	if (!repEl)
	{
		printf("Bad XML format - no repetoire node\n");
		return false;
	}

	const char * repname = repEl->Attribute("name");
	if (!skelname)
	{
		printf("Bad XML format - no repetoire name\n");
		return false;
	}

	if (!Repetoire::LoadRepetoire(repname,repname))
	{
		printf("Error loadinf repetoire %s\n",repname);
		return false;
	}
	rep = Repetoire::GetByName(repname);


	if (!skinEl)
	{
		printf("Bad XML format - no skin node\n");
		return false;
	}
	const char * skinname = skinEl->Attribute("name");

	if (!skinEl)
	{
		printf("Bad XML format - no skin name\n");
		return false;
	}


	Material * mat = Material::CreateManagedMaterial(skinname);
	if (mat)
	{
		mat->setAmbient(1,1,1);
		mat->setTexture(skinname,true);
	}
	
	SetMaterial(skinname);

	return true;
}


bool 
Actor::SetSkel(const char * name)
{
 	if (skelSet)
		SkelChar::RetireInstance(skelName.c_str(),skelInst);

	skelName = name;
	skelInst = SkelChar::CreateInstance(name);
	if (skelInst >= 0)
	{
		skelSet = true;
		SkelCharInst * sk = SkelChar::GetInstancePtr(skelName.c_str(),skelInst);
		savedPose = Pose(sk->GetNumJoints());
		currentPose = Pose(sk->GetNumJoints());
		SetPose(currentPose);
		AddChild(sk);
	}
	return (skelInst >= 0);
}

void 
Actor::SetPose(Pose & p)
{
	SkelCharInst * sk = SkelChar::GetInstancePtr(skelName.c_str(),skelInst);
	if (p.GetSize() != sk->GetNumJoints())
		return;
	currentPose = p;
	sk->SetPose(currentPose);
	sk->Deform();
}

void 
Actor::SetHeading(float h)
{
	heading = h;
}

bool 
Actor::PlayAnim(const char * name)
{
	if (!skelSet)
		return false;
	SkelCharInst * sk = SkelChar::GetInstancePtr(skelName.c_str(),skelInst);

	a1.startTime = TimeStamp();
	a1.name = name;
	a1.startxform = xform;
	a1.startxform.Transpose();
	a1.starthead = heading;
	a1.startPos = Pos;
	playing = true;
		
	//save the current pose
	savedPose = currentPose;
//	printf("savedPose set: %f\n",savedPose.GetHeading());

	playing = true;
	stillBlend = true;
	return true;
}

void 
Actor::SetMaterial(const char * name)
{
	matName = name;
}

float 
Actor::GetHeading()
{
	return heading;
}

void 
Actor::Update()
{
	if (!skelSet)
		return;

	//if we're not playing an animation, there's nothing to do!
	if (!playing)
		return;

	Pose p1;
	TimeStamp now = TimeStamp();

	KeyFrameAnim * kf1;
	float elapsed1;

	SkelCharInst * sk = SkelChar::GetInstancePtr(skelName.c_str(),skelInst);

	//Get pose for anim1
	kf1 = KeyFrameAnim::GetPtr(a1.name.c_str());
	elapsed1 = (float)a1.startTime.Interval(now);
	kf1->SetTime(elapsed1*1000);
	p1 = kf1->GetPose();


	//figure out our interpolation value for the active blend
	float slerp = 0.0f;
	
	if (stillBlend)
	{
		float offset = (float)a1.startTime.Interval(now)*1000.0f;
		slerp = (offset) / defaultWindow;
	}

	p1.SetInitialTransform(a1.startPos,a1.starthead);

	if (stillBlend)
		currentPose.LinearInterpolate(savedPose,p1,slerp);
	else 
		currentPose = p1;
	

	//check if done blending!
	if (slerp >= 1)
		stillBlend = false;

	// check if done animation 1 - wait for blend to end, then stop
	if (elapsed1*1000 >= kf1->GetDuration())
	{
		if (!stillBlend)
			playing = false;
	}

	//apply the current pose, and deform the skeleton
	sk->SetPose(currentPose);

	

	heading = currentPose.GetHeading();
	Angle.setAngles(0,heading,0);
//	printf("SetAngle:%.2f\n",currentPose.GetHeading());
	
	currentPose.GetPosition(Pos);
	
	sk->Deform();
	
}

bool
Actor::IsIdle()
{
	return !playing;
}
void 
Actor::GetJointXform(const char * name, Matrix4x4 & m)
{
	SkelCharInst * sk = SkelChar::GetInstancePtr(skelName.c_str(),skelInst);
	sk->GetJointXform(name,m);
}

bool 
Actor::PlayAction(const char * actionname)
{
	if (!rep)
		return false;
	
	std::string animName = rep->PickAnim(actionname,attributes);
	if (animName == "")
		return false;

	PlayAnim(animName.c_str());
	return true;

}

void 
Actor::SetAttribute(const char * name, float val)
{
	attributes[name] = val;
}

std::string 
Actor::GetCurrentAnimation()
{
	if (playing)
		return a1.name;
	return "no animation";
}

float
Actor::GetAttribute(const char * name)
{
	return attributes[name];
}
