#define REQUIRE_IOSTREAM
#include <maya/MStatus.h>
#include <maya/MString.h> 
#include <maya/MFileIO.h>
#include <maya/MGlobal.h>
#include <maya/MLibrary.h>
#include <maya/MIOStream.h>
#include <maya/MItDag.h>
#include <maya/MDagPath.h>
#include <maya/MPointArray.h>
#include <maya/MFnDagNode.h>
#include <maya/MFnMesh.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItDependencyNodes.h>
#include <maya/MItDependencyGraph.h>
#include <maya/MFnSkinCluster.h>
#include <maya/MDagPathArray.h>
#include <maya/MItGeometry.h>
#include <maya/MFnIKJoint.h>
#include <maya/MEulerRotation.h>
#include <maya/MTime.h>
#include <maya/MFnAnimCurve.h>
#include <maya/MAnimControl.h>
#include <set>;
#include <map>;
#include <matrix4x4.h>
#include <vector3.h>

#define MAXBONES 4

using namespace std;


typedef struct 
{
	float rot[3];
}xform;

bool cmpXform(xform & x1, xform & x2)
{
	return ( (x1.rot[0] == x2.rot[0]) 
		   &&(x1.rot[1] == x2.rot[1]) 
		   &&(x1.rot[2] == x2.rot[2])); 
}
struct joint
{
	char name[20];
	float relpos[3];
	float orient[3];
	int parent;
	//Matrix4x4 xform;
	MString mj;
	std::map<float,xform> keys;
};

struct vertex
{
	float pos[3];
	float tex[2];
	int bone[MAXBONES];
	float weights[MAXBONES];
	int numbones;
};

struct triangle
{
	int norms[3];
	int verts[3];
};

struct normal
{
	float orient[3];
};

struct skelchar
{
	normal * norms;
	vertex * verts;
	triangle * polys;
	joint * joints;
	int numverts;
	int numpolys;
	int numjoints;
	int numnorms;
};

std::set<float>  findFrames(float mintime, float maxtime)
{
	std::set<float> keytimes;
	keytimes.insert(mintime);
	keytimes.insert(maxtime);
	MStatus stat;
	MItDag dagIter( MItDag::kBreadthFirst, MFn::kInvalid, &stat );
    MTime maxKeyTime;
	maxKeyTime.setValue(0);

   for ( ; !dagIter.isDone(); dagIter.next())
   {
      MDagPath dagPath;
      stat = dagIter.getPath( dagPath );

      MFnDagNode dn( dagPath );

      // find all the motion nodes
      MItDependencyGraph dgIter( dagPath.node(),
                                 MFn::kAnimCurve,
                                 MItDependencyGraph::kUpstream,
                                 MItDependencyGraph::kBreadthFirst,
                                 MItDependencyGraph::kNodeLevel,
                                 &stat );
      if ( stat )
      {
         // get all the keyframe times
         for ( ; !dgIter.isDone(); dgIter.next())
         {
            MObject anim = dgIter.thisNode( &stat );
            MFnAnimCurve animCurve( anim, &stat );

            // found some keyframe information
            if ( MS::kSuccess == stat )
            {
               unsigned int numKeys = animCurve.numKeyframes( &stat );
               for ( unsigned int currKey = 0; currKey < numKeys; currKey++ )
               {
                  // truncating values here...
                  MTime keytime = animCurve.time( currKey, &stat );
				  if ((keytime.as(MTime::kMilliseconds) >= mintime) && (keytime.as(MTime::kMilliseconds)<= maxtime))
					keytimes.insert(keytime.as(MTime::kMilliseconds));
					//printf("key at %.2fms\n",keytime.as(MTime::kMilliseconds));
                  // store this key frame time somewhere.
                  //storekeytime( keytime );

                  // and store the maximum frame time
                  //if ( time > maxKeyTime )
                    // maxKeyTime = time;
               }
            }
         }
      }
   }

	return keytimes;
}

void writeAnim(skelchar * skel, std::set<float> times, const char * fname)
{
	FILE * f = fopen(fname,"wb");
	MAnimControl animctrl;
	
	float minTime = 10000000;

	for(std::set<float>::iterator i = times.begin(); i != times.end(); i++)
	{
		MTime time;
		time.setUnit(MTime::kMilliseconds);
		time.setValue(*i);
		animctrl.setCurrentTime(time);
		
		minTime = min(time.as(MTime::kMilliseconds),minTime);
		
		
		for(int j = 0; j < skel->numjoints; j++)
		{
			MSelectionList select;
			MGlobal::selectByName(skel->joints[j].mj,	MGlobal::kReplaceList);
			MGlobal::getActiveSelectionList(select);

			MObject jnt;
			select.getDependNode(0,jnt);
			
			MFnIkJoint mj(jnt);
			MEulerRotation rot;
			mj.getRotation(rot);
			xform x;
			x.rot[0] = rot.x;
			x.rot[1] = rot.y;
			x.rot[2] = rot.z;
			skel->joints[j].keys[time.as(MTime::kMilliseconds)] = x;

		}
	}

	int numkeys = 0;
	//recopy list without duplicates
	for (int j = 0; j < skel->numjoints; j++)
	{
		xform lastx;
		std::map<float,xform> tempmap;
		tempmap.clear();
	
		for (std::map<float,xform>::iterator i = skel->joints[j].keys.begin(); i != skel->joints[j].keys.end(); i++)
		{
			float time = i->first;
			xform & x = i->second;
			if (i == skel->joints[j].keys.begin() || !cmpXform(lastx,x) )
			{
				numkeys++;
				tempmap[time] = x;
				//printf("joint %s\ttime %f\t %.2f,%.2f %.2f\n",skel->joints[j].name,time,x.rot[0],x.rot[1],x.rot[2]);
				//lastx = x;
				memcpy(&lastx,&x,sizeof(xform));
			}
				

		}
		skel->joints[j].keys = tempmap;
	}

	fwrite(&numkeys,1,4,f);
	fwrite(&skel->numjoints,1,4,f);

	for (int j = 0; j < skel->numjoints; j++)
	{
		for (std::map<float,xform>::iterator i = skel->joints[j].keys.begin(); i != skel->joints[j].keys.end(); i++)
		{
			//joint num
			fwrite(&j,4,1,f);
			float time = i->first - minTime;
			xform & x = i->second;

			//time
			fwrite(&time,4,1,f);
			
			//rotation
			fwrite(&x.rot,4,3,f);
		}
		


	}
	fclose(f);
}

void writeSkel(const char * filename, skelchar * skel)
{//sz cnt
	FILE * f = fopen(filename,"wb");
	fwrite("SKEL",1,5,f);
	fwrite(&skel->numjoints,4,1,f);
	fwrite(&skel->numverts,4,1,f);
	fwrite(&skel->numpolys,4,1,f);
	fwrite(&skel->numnorms,4,1,f);

	for (int i = 0; i < skel->numjoints;i++)
	{
		fwrite(skel->joints[i].name,1,21,f);
		fwrite(skel->joints[i].relpos,4,3,f);
		fwrite(skel->joints[i].orient,4,3,f);
		fwrite(&skel->joints[i].parent,4,1,f);
	}

	for (int i = 0; i < skel->numverts;i++)
	{
		fwrite(skel->verts[i].pos,4,3,f);
		fwrite(&skel->verts[i].numbones,4,1,f);
		fwrite(skel->verts[i].bone,4,skel->verts[i].numbones,f);
		fwrite(skel->verts[i].weights,4,skel->verts[i].numbones,f);
	}

	for (int i = 0; i < skel->numpolys;i++)
	{
		fwrite(skel->polys[i].verts,4,3,f);
		fwrite(skel->polys[i].norms,4,3,f);
	}

	for (int i = 0; i < skel->numnorms;i++)
	{
		fwrite(skel->norms[i].orient,4,3,f);
	}

	fclose(f);

}

void fillJointParams(MFnSkinCluster & skin, MFnIkJoint& j, joint * thejoint)
{
	MStatus stat;
	strncpy(thejoint->name,j.name().asChar(),sizeof(thejoint->name)-1);
	MDagPath path = MDagPath::getAPathTo(j.parent(0));
	int pindex = skin.indexForInfluenceObject(path,&stat);
	if (stat == MStatus::kSuccess)
		thejoint->parent = pindex;
	else
		thejoint->parent = -1;

	MEulerRotation er;
	j.getOrientation(er);
	
	MVector rot = er.asVector();
	thejoint->orient[0] = (float)rot.x;
	thejoint->orient[1] = (float)rot.y;
	thejoint->orient[2] = (float)rot.z;
	
	//MVector pos =j.translation(MSpace::kPreTransform);
	MVector pos =j.translation(MSpace::kTransform);
	thejoint->relpos[0] = (float)pos.x;
	thejoint->relpos[1] = (float)pos.y;
	thejoint->relpos[2] = (float)pos.z;

}

MObject getSkinClusterForJoint(MObject & j)
{
	MStatus stat;
	MItDependencyNodes itNodes(MFn::kSkinClusterFilter);
    for(/**/;!itNodes.isDone();itNodes.next())
	{
		MFnSkinCluster fnSkin(itNodes.item());
		
		MDagPathArray infObjArray;
		fnSkin.influenceObjects(infObjArray,&stat);
		for(uint i = 0; i<infObjArray.length();i++)
			if (infObjArray[i].node() == j)
				return itNodes.item();
		
	}
	return MObject::kNullObj;
}

skelchar * loadSkeleton(const char * rootname)
{
	//find the object with rootname as its name
	MSelectionList select;
	MGlobal::selectByName(rootname,	MGlobal::kReplaceList);
	MGlobal::getActiveSelectionList(select);
	if (select.isEmpty())
	{
		cout << "No object " << rootname << " in scene" << endl;
		return NULL;
	}

	//make sure rootname is a joint
	MObject root;
	select.getDependNode(0,root);
	if (!root.hasFn(MFn::kJoint))
	{
		cout << "Object " << rootname << " is not a joint" << endl;
		return NULL;
	}

	//get the skincluster object associated with the joint
	MFnIkJoint rj(root);
	MObject skinClusterObj = getSkinClusterForJoint(root);
	if (skinClusterObj.isNull())
	{
		cout << "Skeleton not bound to geometry" << endl;
		return NULL;
	}
	MFnSkinCluster fnSkin(skinClusterObj);

	//joint * rootJoint = new(joint);
	//fillJointParams(rj,rootJoint);
	//rootJoint->parent = NULL;
	//getchildjoints(rj,rootJoint);
	MStatus stat;

	//get the full set of joints
	MDagPathArray infObjArray;
    fnSkin.influenceObjects(infObjArray,&stat);
	

	skelchar * skel = new(skelchar);

	//get the mesh object
	MObjectArray meshArray;
	fnSkin.getInputGeometry(meshArray);
	MFnMesh fnMesh(meshArray[0]);


	//load the poly info
	skel->numpolys = fnMesh.numPolygons();
	skel->polys = new(triangle[skel->numpolys]);
	
	for (unsigned int i = 0; i < skel->numpolys; i++)
	{
		MIntArray vertlist;
		MIntArray norms;
		fnMesh.getFaceNormalIds(i,norms);
		fnMesh.getPolygonVertices(i,vertlist);
		if (vertlist.length() != 3)
		{
			cout << "Non Triangle Found" << endl;
			delete(skel->polys);
			delete(skel);			
			return NULL;
		}
		skel->polys[i].verts[0] = vertlist[0];
		skel->polys[i].verts[1] = vertlist[1];
		skel->polys[i].verts[2] = vertlist[2];
		skel->polys[i].norms[0] = norms[0];
		skel->polys[i].norms[1] = norms[1];
		skel->polys[i].norms[2] = norms[2];
	}

	//load the normal info
	skel->numnorms = fnMesh.numNormals();
	MFloatVectorArray norms;
	fnMesh.getNormals(norms);
	skel->norms = new(normal[skel->numnorms]);
	for (unsigned int i = 0; i < skel->numnorms; i++)
	{
		skel->norms[i].orient[0] = norms[i].x;
		skel->norms[i].orient[1] = norms[i].y;
		skel->norms[i].orient[2] = norms[i].z;
	}

	//load the vertex info
	MPointArray vertexList;
	fnMesh.getPoints( vertexList, MSpace::kLast);
	skel->verts = new(vertex[vertexList.length()]);
	skel->numverts = vertexList.length();

	for (unsigned int i = 0; i < vertexList.length(); i++)
	{
		vertexList[i].cartesianize();
		
		skel->verts[i].pos[0] = vertexList[i].x;
		skel->verts[i].pos[1] = vertexList[i].y;
		skel->verts[i].pos[2] = vertexList[i].z;
		skel->verts[i].tex[0] = 0.0f;
		skel->verts[i].tex[1] = 0.0f;
		skel->verts[i].numbones = 0.0f;
	}



	skel->joints = new (joint[infObjArray.length()]);
	skel->numjoints = infObjArray.length();

	//loop through all joints
	for(uint i = 0; i<infObjArray.length();i++)
	{        
		//fill out the joint structure
		MFnIkJoint jnt(infObjArray[i]);
		fillJointParams(fnSkin,jnt,&skel->joints[i]);
		skel->joints[i].mj = jnt.fullPathName();


		
		MSelectionList result;
        MFloatArray weights;
        fnSkin.getPointsAffectedByInfluence (infObjArray[i], result,  weights);
		
		MDagPath mesh;
		MObject vertices;
		result.getDagPath(0,mesh,vertices);
		MItGeometry itGeom(mesh,vertices);
		int len = weights.length();
		for(uint k = 0;!itGeom.isDone();itGeom.next(),k++)
		{
			int index = itGeom.index();
			skel->verts[index].bone[skel->verts[index].numbones] = i;
			skel->verts[index].weights[skel->verts[index].numbones] = weights[k];
			skel->verts[index].numbones++;
			if (skel->verts[index].numbones > MAXBONES)
			{
				cout << "too many joints per vertex!" << endl;
				//memory leak! wheeee!
				return NULL;
			}
		}
	}


	return skel;
}


// initialize Maya, before any Maya operations are performed.
// You might do this in main for example:
int main(int narg, char** carg)
{
   MStatus stat = MLibrary::initialize ( "myExporter" );
   if ( !stat )
      return false;

   // figure out the file name to open
   char* fileName = "c:/temp/cube.mb";

   // prepare Maya to read a new scene file
   MFileIO::newFile( true );

   // read the scene file
   stat = MFileIO::open( fileName );
   if ( !stat )
      return false;

   // remove any temporary data created by Maya while loading
   // things like the undo list which we won't be using
   stat = MGlobal::executeCommand( "delete -ch" );
   if ( !stat )
      return false;

	
   std::set<float> times = findFrames(0,1000);
	skelchar * skel = loadSkeleton("root");
	if (!skel)
	{
		printf("error loading maya model");
		return 1;
	}
	writeSkel("test.skel",skel);
	writeAnim(skel,times,"test.anim");

	// now shut down Maya, or if you want to process another file,
   // just make another call to MFileIO::newFile(), and MFileIO::open()
   MLibrary::cleanup();
   return 1;
}

